#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <psapi.h>
#include <time.h>
#include <assert.h>
#include <vector>

#include <pdh.h>
#include <pdhmsg.h>

//------------------------------------------------------------------------------------------------------------------
// getcpuload()
//   directly prints the CPU usage on screen. This function need to be called twice with a minimum of 1 seconds
//   delay (msdn guideline) to display something usefull.
//   Also returns the usage in percent 0-100 where 100 means the system is working at maximum capacity.
//   Note for multiprocessor systems:
//   If one CPU is working at max capacity the result will (if using (_total) for PdhAddCounter() ) show a maximum
//   workload of 50% unless the other CPU(s) is also working at max load. 
//------------------------------------------------------------------------------------------------------------------
INT getcpuload()
{
	static HQUERY h_query;
	static HCOUNTER h_counter;

	static bool b_runonce = true;
	PDH_STATUS n_status;
	if(b_runonce) {
		if((n_status = PdhOpenQuery(NULL, 0, &h_query)) != ERROR_SUCCESS) {
			fprintf(stderr, "PdhOpenQuery() ***Error: 0x%X\n", n_status);
			return 0;
		}
		PdhAddCounter(h_query, TEXT("\\Processor(_Total)\\% Processor Time"), 0, &h_counter);
		// a total of all cpu's in the system

		PdhCollectQueryData(h_query);
		// no error checking here

		b_runonce = false;
		return 0;
	}

	n_status;
	if((n_status = PdhCollectQueryData(h_query)) != ERROR_SUCCESS) {
		fprintf(stderr, "PhdCollectQueryData() ***Error: 0x%X\n",n_status);
		if(n_status==PDH_INVALID_HANDLE) 
			fprintf(stderr, "PDH_INVALID_HANDLE\n");
		else if(n_status==PDH_NO_DATA)
			fprintf(stderr, "PDH_NO_DATA\n");
		else
			fprintf(stderr, "Unknown error\n");
		return 0;
	}

	DWORD n_result;
	PDH_FMT_COUNTERVALUE t_value;
	n_status = PdhGetFormattedCounterValue(h_counter, PDH_FMT_DOUBLE | PDH_FMT_NOCAP100, &n_result, &t_value);
	if(n_status != ERROR_SUCCESS) {
		fprintf(stderr, "PdhGetFormattedCounterValue() ***Error: 0x%X\n",n_status);
		return 0;
	}
	int cput = (int)(t_value.doubleValue + .5);

	printf("CPU usage: %3d%%\n", cput);

	return cput;
}

const char *p_s_CurrentTime()
{
	time_t t_timet = time(NULL);
	tm t_time = *localtime(&t_timet);
	// get actual time

	int n_year = t_time.tm_year + 1900;
	int n_month = t_time.tm_mon + 1;
	int n_day = t_time.tm_mday;
	int n_hours = t_time.tm_hour;
	int n_minutes = t_time.tm_min;
	int n_seconds = t_time.tm_sec;
	static char p_s_buffer[20];
	sprintf(p_s_buffer, "%04d-%02d-%02dT%02d:%02d:%02d",
		n_year, n_month, n_day, n_hours, n_minutes, n_seconds);
	assert(strlen(p_s_buffer) == 19);
	// format time string

	return p_s_buffer;
}

void PrintProcessNameAndID(DWORD n_process_id)
{
    char p_s_process_name[MAX_PATH] = "<unknown>";

    HANDLE h_process = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_SET_INFORMATION |
		PROCESS_VM_READ, FALSE, n_process_id);
    // get a handle to the process.

    if(h_process != NULL) {
        HMODULE h_mod;
        DWORD n_needed;
        if(EnumProcessModules(h_process, &h_mod, sizeof(h_mod), &n_needed))
            GetModuleBaseName(h_process, h_mod, p_s_process_name, sizeof(p_s_process_name) / sizeof(char));
    }
    // get the process name.

	static const char *p_name_list[] = {
		"cl.exe"
	};
	bool b_match = false;
	for(size_t i = 0; i < sizeof(p_name_list) / sizeof(p_name_list[0]); ++ i) {
		if(!strncmp(p_s_process_name, p_name_list[i], strlen(p_name_list[i]))) {
			b_match = true;
			break;
		}
	}
	// tries to match filename

	if(b_match) {
		//printf("%s  (PID: %u)\n", p_s_process_name, n_process_id);
		// print the process name and identifier.

		DWORD n_process_mask, n_system_mask, n_priority;
		if((n_priority = GetPriorityClass(h_process)) != IDLE_PRIORITY_CLASS) {
			if(!SetPriorityClass(h_process, IDLE_PRIORITY_CLASS)) {
				fprintf(stderr, "error: SetPriorityClass() failed\n");
			} else {
				printf("%s - set priority class of \'%s\' (PID %d) to below normal\n",
					p_s_CurrentTime(), p_s_process_name, n_process_id);
			}
		}
		if(GetProcessAffinityMask(h_process, &n_process_mask, &n_system_mask)) {
			//printf("\taffinity mask: 0x%08x\n", n_process_mask);
			if(n_process_mask != n_system_mask) {
				if(!SetProcessAffinityMask(h_process, n_system_mask))
					fprintf(stderr, "error: SetProcessAffinityMask() failed\n");
				else {
					printf("%s - set affinity mask of \'%s\' (PID %d) to 0x%04x\n",
						p_s_CurrentTime(), p_s_process_name, n_process_id, n_system_mask);
				}
			}
		}
	}
	// find seti at home processes

    CloseHandle(h_process);
}

static bool b_stop_service = false;

void DoTheJob()
{
	/*std::vector<int> core_bits, core_occupation;

	DWORD n_system_mask;
	{
		HANDLE h_process = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, GetCurrentProcessId());
		DWORD n_process_mask;
		if(!GetProcessAffinityMask(h_process, &n_process_mask, &n_system_mask)) {
			fprintf(stderr, "error: GetProcessAffinityMask() failed\n");
			return -1;
		}
		CloseHandle(h_process);
	}
	// get system affinity mask

	for(int i = 0; i < 8 * sizeof(DWORD); ++ i) {
		DWORD bit = (DWORD)1 << i;
		if(n_system_mask & bit)
			core_bits.push_back(bit);
	}
	// get processor bits

	core_occupation.clear();
	core_occupation.resize(core_bits.size(), 0);
	// fill core occupation with nulls
	*/ // not needed right now

	/*HWND h_wnd = GetConsoleWindow();
	ShowWindow(h_wnd, SW_HIDE);
	SetWindowLong(h_wnd, GWL_EXSTYLE, GetWindowLong(h_wnd, GWL_EXSTYLE) | WS_EX_TOOLWINDOW);
	SetWindowPos(h_wnd, HWND_TOP, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOZORDER | SWP_FRAMECHANGED);
	ShowWindow(h_wnd, SW_SHOW);
	// make console be toolwindow, effectively removing it from taskbar ;)*/ // doesn't work

	printf("keeping cl.exe processes scheduled accross all CPU cores and running with below normal priority\n");

	getcpuload();
	// to initialize the performance counter

	for(; !b_stop_service;) {
		Sleep(5000);
		if(getcpuload() < 95)
			continue;
		// get CPU load and only proceed if it is above 95%

		std::vector<DWORD> process_list;
		process_list.resize(100, 0);
		for(;;) {
			DWORD n_process_num;
			if(!EnumProcesses(&process_list[0], process_list.size() * sizeof(DWORD), &n_process_num)) {
				fprintf(stderr, "error: EnumProcesses() failed\n");
				return;
			}
			n_process_num /= sizeof(DWORD);
			if(process_list.size() <= n_process_num)
				process_list.resize(n_process_num * 2 + 1, 0);
			else {
				process_list.resize(n_process_num);
				break;
			}
		};
		// get list of processes

		for(size_t i = 0, n = process_list.size(); i < n; ++ i) {
			if(!process_list[i])
				continue;
			PrintProcessNameAndID(process_list[i]);
		}
		// find cl.exe and set priority / affinity masks
	}

	return;
}

#define SVCNAME "VisualStudioBuildSchedulingService"

static SERVICE_STATUS t_service_status; 
static SERVICE_STATUS_HANDLE h_service_status; 

void ReportStatus(DWORD dwCurrentState, DWORD dwWin32ExitCode, DWORD dwWaitHint)
{
    static DWORD dwCheckPoint = 1;

    t_service_status.dwCurrentState = dwCurrentState;
    t_service_status.dwWin32ExitCode = dwWin32ExitCode;
    t_service_status.dwWaitHint = dwWaitHint;

    if(dwCurrentState == SERVICE_START_PENDING)
        t_service_status.dwControlsAccepted = 0;
    else
		t_service_status.dwControlsAccepted = SERVICE_ACCEPT_STOP;
    if((dwCurrentState == SERVICE_RUNNING) || (dwCurrentState == SERVICE_STOPPED))
        t_service_status.dwCheckPoint = 0;
    else
		t_service_status.dwCheckPoint = dwCheckPoint ++;
    // fill in the SERVICE_STATUS structure

    SetServiceStatus(h_service_status, &t_service_status);
    // report the n_status of the service to the SCM
}

void WINAPI ControlHandler(DWORD n_control)
{
	switch(n_control) {  
	case SERVICE_CONTROL_STOP: 
		ReportStatus(SERVICE_STOP_PENDING, NO_ERROR, 0);

		b_stop_service = true;
		// @note use event object / condition variable if working with threads

		ReportStatus(t_service_status.dwCurrentState, NO_ERROR, 0);
		return;
	case SERVICE_CONTROL_INTERROGATE: 
		break; 
	default: 
		break;
	} 
}

void WINAPI ServiceMain(DWORD n_arg_num, LPSTR *p_arg_list)
{
	if(!(h_service_status = RegisterServiceCtrlHandler(SVCNAME, &ControlHandler)))
		return;
	// @note service n_status handle does not have to be closed

	t_service_status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    t_service_status.dwServiceSpecificExitCode = 0;

	ReportStatus(SERVICE_START_PENDING, NO_ERROR, 1000);

	// @todo initialize the service (have 1000 msec to do so)

	ReportStatus(SERVICE_RUNNING, NO_ERROR, 0);

	DoTheJob();

	ReportStatus(SERVICE_STOPPED, NO_ERROR, 0);
}

int main(int n_arg_num, const char **p_arg_list)
{
#if 0
	DoTheJob(); // debug
#else
	SERVICE_TABLE_ENTRY p_dispatch_table[] = {
        {SVCNAME, &ServiceMain},
        {NULL, NULL}
    };
 
    if(!StartServiceCtrlDispatcher(p_dispatch_table))
		return -1;
#endif // 1
	return 0;
}
